﻿#define _CRT_RAND_S
#include <stdlib.h>
#include <windows.h>
#include <commctrl.h>
#include <shlobj.h>
#include <shlwapi.h>
#include <gl\gl.h>
#pragma hdrstop

#include "Shared.h"
#include "..\Shared\CObeliskException.h"
#include "CApplication.h"
#include "CThreadWrapper.h"
#include "CFileWrapper.h"

using namespace std;
using namespace Obelisk;



IMainController::~IMainController()
{
}



CApplication::CApplication(HINSTANCE aInstance)
{
	mInstance = aInstance;

	mDirectorySet = false;
	mCaptionSet = false;
	mWindowXSize = -1;
	mTimerStarted = false;
	mTimerCallbackActive = false;
	mMouseLeftLast = false;
	mMouseRightLast = false;
	mAltAttached = false;
}



CApplication::~CApplication()
{
}



void CApplication::Run()
{
	mMainController = CreateMainController(this);
	Initialise();
	mMainController->Initialise();
	while (true)
	{
		MSG Message;
		int Result = GetMessage(&Message, NULL, 0, 0);
		Assert(Result != -1)
		if (Result == 0)
			break;
		//TranslateMessage(&Message);
		DispatchMessage(&Message);
	}
	delete mMainController;
}



void CApplication::Initialise()
{
	if (!mCaptionSet)
		throw CObeliskException(SL("Program forgot to call SetCaption()"));
	if (mWindowXSize == -1)
		throw CObeliskException(SL("Program forgot to call SetSize()"));

	mMainThreadID = GetCurrentThreadId();

	HMODULE Kernel32Dll = LoadLibrary(L"Kernel32.dll");
	if (Kernel32Dll == NULL)
		throw CException("LoadLibrary() failed");
	mCancelSynchronousIo = (BOOL (WINAPI *)(HANDLE))GetProcAddress(Kernel32Dll, "CancelSynchronousIo");
	FreeLibrary(Kernel32Dll);

	INITCOMMONCONTROLSEX icc;
	ZeroMemory(&icc, sizeof(INITCOMMONCONTROLSEX));
	icc.dwSize = sizeof(INITCOMMONCONTROLSEX);
	bool ICCResult = InitCommonControlsEx(&icc);
	Assert(ICCResult)

	WNDCLASSEX ClassInfo;
	ZeroMemory(&ClassInfo, sizeof(WNDCLASSEX));
	ClassInfo.cbSize = sizeof(WNDCLASSEX);
	ClassInfo.style = CS_OWNDC;
	ClassInfo.lpfnWndProc = MainWndProc;
	//ClassInfo.cbClsExtra = sizeof(LONG_PTR);
	ClassInfo.hInstance = mInstance;
	ClassInfo.hCursor = LoadCursor(NULL, IDC_ARROW);
	Assert(ClassInfo.hCursor != NULL)
	ClassInfo.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
	ClassInfo.lpszClassName = L"ObeliskMainWindow";
	ATOM RCResult = RegisterClassEx(&ClassInfo);
	Assert(RCResult != 0)

	RECT r;
	bool SRResult = SetRect(&r, 0, 0, mWindowXSize, mWindowYSize);
	Assert(SRResult)
	bool AWRResult = AdjustWindowRect(&r, WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_POPUPWINDOW |
			WS_CAPTION, false);
	Assert(AWRResult)
	int XSize = r.right - r.left;
	int X = (GetSystemMetrics(SM_CXSCREEN) - XSize) / 2;
	int YSize = r.bottom - r.top;
	int Y = (GetSystemMetrics(SM_CYSCREEN) - r.bottom) / 2;
	mWindow = CreateWindowEx(WS_EX_APPWINDOW, L"ObeliskMainWindow", mCaption.GetData(),
			WS_MINIMIZEBOX | WS_MAXIMIZEBOX | WS_POPUPWINDOW | WS_CAPTION | WS_VISIBLE, X, Y,
			XSize, YSize, NULL, NULL, NULL, this);
	Assert(mWindow != NULL)
	SetLastError(0);
	LONG_PTR SWLResult = SetWindowLongPtr(mWindow, GWLP_USERDATA, (LONG_PTR)this);
	Assert(SWLResult != 0 || GetLastError() == 0)

	//ShowCursor(FALSE);



	PIXELFORMATDESCRIPTOR pfd;
	ZeroMemory(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.cColorBits = 24;
	pfd.cDepthBits = 16;
	mDC = GetDC(mWindow);
	Assert(mDC != NULL)
	int PixelFormat = ChoosePixelFormat(mDC, &pfd);
	if (PixelFormat == 0)
		throw CObeliskException(SL("ChoosePixelFormat() failed"));
	// TODO: check that pixel format is good enough
	bool SPFResult = SetPixelFormat(mDC, PixelFormat, &pfd);
	if (!SPFResult)
		throw CException("SetPixelFormat() failed");
	mRenderingContext = wglCreateContext(mDC);
	if (mRenderingContext == NULL)
		throw CException("wglCreateContext() failed");
	mAltRenderingContext = wglCreateContext(mDC);
	if (mAltRenderingContext == NULL)
		throw CException("wglCreateContext() failed (1)");
	bool SLResult = wglShareLists(mRenderingContext, mAltRenderingContext);
	Assert(SLResult)
	bool WMCResult = wglMakeCurrent(mDC, mRenderingContext);
	Assert(WMCResult)
	Assert(glGetError() == GL_NO_ERROR)



	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glClearDepth(1.0f);
	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	glShadeModel(GL_SMOOTH);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	SetSpriteViewport(mWindowXSize, mWindowYSize);
	Assert(glGetError() == GL_NO_ERROR)
}



LRESULT CALLBACK CApplication::MainWndProc(HWND hwnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	SetLastError(0);
	LONG_PTR UserData = GetWindowLongPtr(hwnd, GWLP_USERDATA);
	Assert(UserData != 0 || GetLastError() == 0)
	CApplication *Application = (CApplication*)UserData;

	switch (uMsg)
	{
	case WM_USER:
		Application->mMainController->OnNotify(wParam, lParam);
		return 0;
	case WM_USER + 1:
		Application->ThreadFinished((IThread*)wParam);
		return 0;
	case WM_USER + 2:
		throw CException("Thread crashed");
	case WM_USER + 3:
	{
		wchar_t *Message = (wchar_t*)wParam;
		CString ExceptionMessage(wcslen(Message), Message);
		delete [] Message;
		throw CObeliskException(ExceptionMessage);
	}
	case WM_CLOSE:
		DestroyWindow(hwnd);
		return 0;
	case WM_DESTROY:
		Application->WindowDestroyed();
		return 0;
	case WM_TIMER:
		if (Application->mTimerStarted)
			Application->OnTimer();
		return 0;
	case WM_KEYDOWN:
		Application->OnKeyDown(wParam);
		return 0;
	case WM_KEYUP:
		Application->OnKeyUp(wParam);
		return 0;
	case WM_SETCURSOR:
		if (Application->OnSetCursor())
			return TRUE;
		break;
	case WM_SETFOCUS:
		if (Application != NULL)
			Application->mMainController->OnActivate();
		break;
	case WM_KILLFOCUS:
		Application->mMainController->OnDeactivate();
		break;
	}
	return DefWindowProc(hwnd, uMsg, wParam, lParam);
}



void CApplication::WindowDestroyed()
{
	if (mTimerStarted)
		StopTimer();
	for (TThreadMap::const_iterator i = mThreads.begin(); i != mThreads.end(); i++)
	{
		i->second->Stop();
		delete i->second;
	}
	bool Result = wglMakeCurrent(NULL, NULL);
	Assert(Result)
	Result = wglDeleteContext(mRenderingContext);
	Assert(Result)
	Result = wglDeleteContext(mAltRenderingContext);
	Assert(Result)
	::PlaySound(NULL, NULL, 0);
	vector<TByte*> Sounds = mSounds.GetValues();
	for (int i = 0; i < Sounds.size(); i++)
		delete Sounds[i];
	PostQuitMessage(0);
}



bool CApplication::OnSetCursor()
{
	int x, y;
	GetCursorPos(x, y);
	RECT Client;
	bool GCRResult = GetClientRect(mWindow, &Client);
	Assert(GCRResult)
	if (x >= 0 && x < Client.right - Client.left && y >= 0 && y < Client.bottom - Client.top)
	{
		SetCursor(NULL);
		return true;
	}
	return false;
}



CString CApplication::GetDataFilePath(CString aFileName)
{
	DWORD Size = MAX_PATH;
	wchar_t *FilePath = new wchar_t [Size + 1];
	while (true)
	{
		DWORD Result = GetModuleFileName(NULL, FilePath, Size + 1);
		Assert(Result != 0 && Result <= Size)
		if (Result == Size)
		{
			delete [] FilePath;
			Size *= 2;
			FilePath = new wchar_t [Size + 1];
		}
		else
			break;
	}
	bool PRFSResult = PathRemoveFileSpec(FilePath);
	Assert(PRFSResult)
	CString DataFilePath = CString(wcslen(FilePath), FilePath) + SL("\\Data\\") + aFileName;
	delete [] FilePath;
	return DataFilePath;
}



bool CApplication::GetUserFilePath(CString aFileName, bool aCreatePath, CString& aPath)
{
	if (!mDirectorySet)
		throw CObeliskException(SL("Program forgot to call SetDirectory()"));
	wchar_t AppDataPath[MAX_PATH];
	HRESULT GFPResult = SHGetFolderPath(NULL, CSIDL_APPDATA, NULL, SHGFP_TYPE_CURRENT,
			AppDataPath);
	if (GFPResult == E_FAIL || GFPResult == 0x80070002)
		return false;
	CString Path = CString(wcslen(AppDataPath), AppDataPath);
	if (Path.GetChar(Path.Last()) != SL("\\"))
		Path += SL("\\");
	Path += mDirectory;
	if (aCreatePath)
	{
		SetLastError(0);
		DWORD Attributes = GetFileAttributes(Path.GetData());
		if (Attributes == INVALID_FILE_ATTRIBUTES)
		{
			Assert(GetLastError() == ERROR_FILE_NOT_FOUND)
			if (!CreateDirectory(Path.GetData(), NULL))
				return false;
		}
		else if ((Attributes & FILE_ATTRIBUTE_DIRECTORY) == 0)
			return false;
	}
	aPath = Path + SL("\\") + aFileName;
	return true;
}



bool CApplication::CancelSynchronousIo(HANDLE aThread)
{
	if (mCancelSynchronousIo == NULL)
		return false;
	return mCancelSynchronousIo(aThread);
}



void CApplication::SetDirectory(CString aDirectory)
{
	Assert(!mDirectorySet);   // TODO: disallow \ and /
	mDirectorySet = true;
	mDirectory = aDirectory;
}



void CApplication::SetCaption(CString aCaption)
{
	Assert(!mCaptionSet)
	mCaptionSet = true;
	mCaption = aCaption;
}



void CApplication::SetSize(int aXSize, int aYSize)
{
	Assert(mWindowXSize == -1 && aXSize > 0 && aYSize > 0)
	mWindowXSize = aXSize;
	mWindowYSize = aYSize;
}



CString CApplication::IntToString(int aNumber, int aLeading0s)
{
	bool Negative;
	if (aNumber >= 0)
		Negative = false;
	else
	{
		Negative = true;
		aNumber = -aNumber;
	}
	int Next = 10;
	int Length = 1;
	while (Next <= aNumber)
	{
		Next *= 10;
		Length++;
	}
	int NumLength = Max(Length, aLeading0s);
	int StrLength = Negative ? 1 + NumLength : NumLength;
	wchar_t *Str = new wchar_t [StrLength];
	int Offset;
	if (!Negative)
		Offset = 0;
	else
	{
		Str[0] = L'-';
		Offset = 1;
	}
	for (int i = 0; i < aLeading0s - Length; i++)
		Str[Offset++] = L'0';
	for (int i = 0; i < Length; i++)
	{
		Next /= 10;
		Str[Offset + i] = ((aNumber / Next) % 10) + L'0';
	}
	return CString(StrLength, Str);
}



void CApplication::RunThread(IThread* aThread)
{
	CThreadWrapper *Thread = new CThreadWrapper(this, aThread);
	Thread->Run();
	mThreads[aThread] = Thread;
}



void CApplication::ThreadFinished(IThread* aThread)
{
	Assert(mThreads.count(aThread) == 1)

	mAltAttached = false;

	mMainController->ThreadFinished(aThread);
	delete mThreads[aThread];
	mThreads.erase(aThread);
}



void CApplication::NotifyWindow(int aMessage, int aData)
{
	bool PMResult = PostMessage(mWindow, WM_USER, aMessage, aData);
	Assert(PMResult)
}



void CApplication::StartTimer()
{
	Assert(!mTimerStarted)
	UINT_PTR Result = SetTimer(mWindow, 1, 20, NULL);
	Assert(Result != 0)
	mTimerStarted = true;
}



void CApplication::StopTimer()
{
	Assert(mTimerStarted)
	bool Result = KillTimer(mWindow, 1);
	Assert(Result)
	mTimerStarted = false;
}



void CApplication::OnTimer()
{
	mMouseLeft = (GetAsyncKeyState(VK_LBUTTON) & ~1) != 0;
	mMouseRight = (GetAsyncKeyState(VK_RBUTTON) & ~1) != 0;
	mTimerCallbackActive = true;
	mMainController->OnTimer();
	mTimerCallbackActive = false;
	mMouseLeftLast = mMouseLeft;
	mMouseRightLast = mMouseRight;
}



unsigned int CApplication::CreateTimer()
{
	return GetTickCount();
}



unsigned int CApplication::Duration(unsigned int aTimer)
{
	return GetTickCount() - aTimer;
}



bool CApplication::MouseButtonDown(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? mMouseRight : mMouseLeft;
}



bool CApplication::MouseButtonClicked(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? mMouseRight && !mMouseRightLast : mMouseLeft && !mMouseLeftLast;
}



bool CApplication::MouseButtonReleased(bool aRight)
{
	Assert(mTimerCallbackActive)
	return aRight ? !mMouseRight && mMouseRightLast : !mMouseLeft && mMouseLeftLast;
}



void CApplication::GetCursorPos(int& aCursorX, int& aCursorY)
{
	POINT Cursor;
	bool GCPResult = ::GetCursorPos(&Cursor);
	Assert(GCPResult)
	bool STCResult = ScreenToClient(mWindow, &Cursor);
	Assert(STCResult)
	aCursorX = Cursor.x;
	aCursorY = Cursor.y;
}



void CApplication::OnKeyDown(int aVirtualKey)
{
	EKey Key;
	if (GetKey(aVirtualKey, Key))
		mMainController->OnKeyDown(Key);
}



void CApplication::OnKeyUp(int aVirtualKey)
{
	EKey Key;
	if (GetKey(aVirtualKey, Key))
		mMainController->OnKeyUp(Key);
}



bool CApplication::IsKeyDown(EKey aKey)
{
	return GetFocus() != NULL && (GetAsyncKeyState(GetVirtualKey(aKey)) & ~1) != 0;
}



bool CApplication::GetKey(int aVirtualKey, EKey& aKey)
{
	switch (aVirtualKey)
	{
	case VK_LEFT:
		aKey = kLeft;
		return true;
	case VK_RIGHT:
		aKey = kRight;
		return true;
	case VK_UP:
		aKey = kUp;
		return true;
	case VK_DOWN:
		aKey = kDown;
		return true;
	case VK_RETURN:
		aKey = kEnter;
		return true;
	case VK_SPACE:
		aKey = kSpace;
		return true;
	case VK_ESCAPE:
		aKey = kEscape;
		return true;
	case VK_BACK:
		aKey = kBackspace;
		return true;
	case VK_TAB:
		aKey = kTab;
		return true;
	case VK_SHIFT:
	case VK_LSHIFT:
		aKey = kLeftShift;
		return true;
	case VK_RSHIFT:
		aKey = kRightShift;
		return true;
	case VK_CONTROL:
	case VK_LCONTROL:
		aKey = kLeftControl;
		return true;
	case VK_RCONTROL:
		aKey = kRightControl;
		return true;
	case VK_MENU:
	case VK_LMENU:
		aKey = kLeftAlt;
		return true;
	case VK_RMENU:
		aKey = kRightAlt;
		return true;
	case (int)'A':
		aKey = kA;
		return true;
	case (int)'B':
		aKey = kB;
		return true;
	case (int)'C':
		aKey = kC;
		return true;
	case (int)'D':
		aKey = kD;
		return true;
	case (int)'E':
		aKey = kE;
		return true;
	case (int)'F':
		aKey = kF;
		return true;
	case (int)'G':
		aKey = kG;
		return true;
	case (int)'H':
		aKey = kH;
		return true;
	case (int)'I':
		aKey = kI;
		return true;
	case (int)'J':
		aKey = kJ;
		return true;
	case (int)'K':
		aKey = kK;
		return true;
	case (int)'L':
		aKey = kL;
		return true;
	case (int)'M':
		aKey = kM;
		return true;
	case (int)'N':
		aKey = kN;
		return true;
	case (int)'O':
		aKey = kO;
		return true;
	case (int)'P':
		aKey = kP;
		return true;
	case (int)'Q':
		aKey = kQ;
		return true;
	case (int)'R':
		aKey = kR;
		return true;
	case (int)'S':
		aKey = kS;
		return true;
	case (int)'T':
		aKey = kT;
		return true;
	case (int)'U':
		aKey = kU;
		return true;
	case (int)'V':
		aKey = kV;
		return true;
	case (int)'W':
		aKey = kW;
		return true;
	case (int)'X':
		aKey = kX;
		return true;
	case (int)'Y':
		aKey = kY;
		return true;
	case (int)'Z':
		aKey = kZ;
		return true;
	case (int)'0':
		aKey = k0;
		return true;
	case (int)'1':
		aKey = k1;
		return true;
	case (int)'2':
		aKey = k2;
		return true;
	case (int)'3':
		aKey = k3;
		return true;
	case (int)'4':
		aKey = k4;
		return true;
	case (int)'5':
		aKey = k5;
		return true;
	case (int)'6':
		aKey = k6;
		return true;
	case (int)'7':
		aKey = k7;
		return true;
	case (int)'8':
		aKey = k8;
		return true;
	case (int)'9':
		aKey = k9;
		return true;
	default:
		return false;
	}
}



int CApplication::GetVirtualKey(EKey aKey)
{
	switch (aKey)
	{
	case kLeft:
		return VK_LEFT;
	case kRight:
		return VK_RIGHT;
	case kUp:
		return VK_UP;
	case kDown:
		return VK_DOWN;
	case kEnter:
		return VK_RETURN;
	case kSpace:
		return VK_SPACE;
	case kEscape:
		return VK_ESCAPE;
	case kBackspace:
		return VK_BACK;
	case kTab:
		return VK_TAB;
	case kLeftShift:
		return VK_LSHIFT;
	case kRightShift:
		return VK_RSHIFT;
	case kLeftControl:
		return VK_LCONTROL;
	case kRightControl:
		return VK_RCONTROL;
	case kLeftAlt:
		return VK_LMENU;
	case kRightAlt:
		return VK_RMENU;
	case kA:
		return (int)'A';
	case kB:
		return (int)'B';
	case kC:
		return (int)'C';
	case kD:
		return (int)'D';
	case kE:
		return (int)'E';
	case kF:
		return (int)'F';
	case kG:
		return (int)'G';
	case kH:
		return (int)'H';
	case kI:
		return (int)'I';
	case kJ:
		return (int)'J';
	case kK:
		return (int)'K';
	case kL:
		return (int)'L';
	case kM:
		return (int)'M';
	case kN:
		return (int)'N';
	case kO:
		return (int)'O';
	case kP:
		return (int)'P';
	case kQ:
		return (int)'Q';
	case kR:
		return (int)'R';
	case kS:
		return (int)'S';
	case kT:
		return (int)'T';
	case kU:
		return (int)'U';
	case kV:
		return (int)'V';
	case kW:
		return (int)'W';
	case kX:
		return (int)'X';
	case kY:
		return (int)'Y';
	case kZ:
		return (int)'Z';
	case k0:
		return (int)'0';
	case k1:
		return (int)'1';
	case k2:
		return (int)'2';
	case k3:
		return (int)'3';
	case k4:
		return (int)'4';
	case k5:
		return (int)'5';
	case k6:
		return (int)'6';
	case k7:
		return (int)'7';
	case k8:
		return (int)'8';
	case k9:
		return (int)'9';
	}
	throw CException("Invalid key");
}



double CApplication::Random()
{
	unsigned int Value;
	errno_t Result = rand_s(&Value);
	Assert(Result == 0)
	return (double)Value / UINT_MAX;
}



int CApplication::Random(int aMin, int aMax)
{
	Assert(aMin <= aMax)
	if (aMin == aMax)
		return aMin;
	unsigned int Value;
	errno_t Result = rand_s(&Value);
	Assert(Result == 0)
	return (Value % (aMax - aMin + 1)) + aMin;
}



void CApplication::LoadTexture(CString aName, bool aTransparent)
{
	CString TexturePath = GetDataFilePath(aName);
	HGDIOBJ hBitmap = (HGDIOBJ)LoadImage(NULL, TexturePath.GetData(), IMAGE_BITMAP, 0, 0,
			LR_LOADFROMFILE | LR_CREATEDIBSECTION);
	if (hBitmap == NULL)
		throw CObeliskException(SL("Image file not found: ") + aName);
	BITMAP Bitmap;
	int GOResult = GetObject(hBitmap, sizeof(BITMAP), &Bitmap);
	Assert(GOResult != 0)
	if (!(Bitmap.bmBitsPixel == 24 || Bitmap.bmBitsPixel == 32))
	{
		bool DOResult = DeleteObject(hBitmap);
		Assert(DOResult)
		throw CException("Unsupported bit depth");
	}
	int XSize = Bitmap.bmWidth;
	int XSize2 = NextPowerOf2(XSize);
	int YSize = Bitmap.bmHeight;
	int YSize2 = NextPowerOf2(YSize);
	unsigned int *Data = new unsigned int [XSize2 * YSize2];
	memset(Data, 0, XSize2 * YSize2 * sizeof(unsigned int));
	for (int y = 0; y < YSize; y++)
	{
		int Offset = Bitmap.bmWidthBytes * (YSize - y - 1);
		for (int x = 0; x < XSize; x++)
		{
			int Mult = Bitmap.bmBitsPixel == 24 ? 3 : 4;
			TByte B = ((TByte*)Bitmap.bmBits)[Offset + x * Mult + 0];
			TByte G = ((TByte*)Bitmap.bmBits)[Offset + x * Mult + 1];
			TByte R = ((TByte*)Bitmap.bmBits)[Offset + x * Mult + 2];
			unsigned int Colour = R | (G << 8) | (B << 16);
			if (aTransparent && Colour == 0xffffff)   // white: transparent
				Data[y * XSize2 + x] = 0x00000000;
			else if (aTransparent && Colour == 0xff00ff)   // magenta: semi-transparent
				Data[y * XSize2 + x] = 0x7f7f7f7f;
			else
				Data[y * XSize2 + x] = 0xff000000 | Colour;
		}
	}
	bool DOResult = DeleteObject(hBitmap);
	Assert(DOResult)

	GLuint Name;
	glGenTextures(1, &Name);
	glBindTexture(GL_TEXTURE_2D, Name);
	glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA8, XSize2, YSize2, 0, GL_RGBA, GL_UNSIGNED_BYTE,
			Data);
	Assert(glGetError() == GL_NO_ERROR)
	mTextures.Add(aName, CTexture(XSize, YSize, (float)XSize / XSize2, (float)YSize / YSize2, Name));
}



int CApplication::NextPowerOf2(int n)
{
	int Next = 1;
	while (n > Next)
		Next <<= 1;
	return Next;
}



CTexture CApplication::GetTexture(CString aName)
{
	return mTextures[aName];
}



void CApplication::SetLinearFiltering(bool aLinearFiltering)
{
	mLinearFiltering = aLinearFiltering;
}



void CApplication::AttachRenderingContext()
{
	#ifndef NO_THREADS
		Assert(wglGetCurrentContext() == NULL && !mAltAttached)
		bool WMCResult = wglMakeCurrent(mDC, mAltRenderingContext);
		Assert(WMCResult)
	#else
		Assert(!mAltAttached)
	#endif
	mAltAttached = true;
}



void CApplication::SetBackgroundColour(float aR, float aG, float aB)
{
	Assert(aR >= 0 && aR <= 1 && aG >= 0 && aG <= 1 && aB >= 0 && aB <= 1)
	glClearColor(aR, aG, aB, 1.0f);
}



void CApplication::BeginFrame()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::EndFrame()
{
	bool SBResult = SwapBuffers(mDC);
	Assert(SBResult);
}



void CApplication::SetSpriteViewport(int aXSize, int aYSize)
{
	GLenum MatrixMode;
	glGetIntegerv(GL_MATRIX_MODE, (GLint*)&MatrixMode);
	glMatrixMode(GL_PROJECTION);
	glTranslatef(-1, 1, 0);
	glScalef(2.0 / aXSize, -2.0 / aYSize, 1);
	glMatrixMode(MatrixMode);
}



void CApplication::DrawSprite(CTexture aTexture, float aX, float aY, float aXSize,
		float aYSize, double aXAlign, double aYAlign)
{
	int XSize = aXSize == -1 ? aTexture.GetXSize() : aXSize;
	int YSize = aYSize == -1 ? aTexture.GetYSize() : aYSize;
	double X = ((double)aX - aXAlign * XSize);
	double Y = ((double)aY - aYAlign * YSize);

	glEnable(GL_TEXTURE_2D);
	glDisable(GL_CULL_FACE);
	glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glBegin(GL_QUADS);
	glTexCoord2f(0, 0);
	glColor3f(1, 1, 1);
	glVertex2f(X, Y);
	glTexCoord2f(aTexture.GetXRatio(), 0);
	glColor3f(1, 1, 1);
	glVertex2f(X + XSize, Y);
	glTexCoord2f(aTexture.GetXRatio(), aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex2f(X + XSize, Y + YSize);
	glTexCoord2f(0, aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex2f(X, Y + YSize);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::TriColour(const CVector& aV0, const CVector& aV1, const CVector& aV2,
		const CColour& aColour)
{
	TriColour(aV0, aColour, aV1, aColour, aV2, aColour);
}

void CApplication::TriColour(const CVector& aV0, const CColour& aC0, const CVector& aV1,
		const CColour& aC1, const CVector& aV2, const CColour& aC2)
{
	glDisable(GL_TEXTURE_2D);
	glEnable(GL_CULL_FACE);
	glBegin(GL_TRIANGLES);
	glColor3f(aC0.R, aC0.G, aC0.B);
	glVertex3f(aV0.X, aV0.Y, aV0.Z);
	glColor3f(aC1.R, aC1.G, aC1.B);
	glVertex3f(aV1.X, aV1.Y, aV1.Z);
	glColor3f(aC2.R, aC2.G, aC2.B);
	glVertex3f(aV2.X, aV2.Y, aV2.Z);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::TriTexture(const CVector& aV0, float aS0, float aT0,
		const CVector& aV1, float aS1, float aT1,
		const CVector& aV2, float aS2, float aT2, CTexture aTexture)
{
	glEnable(GL_TEXTURE_2D);
	glBindTexture(GL_TEXTURE_2D, aTexture.GetTextureName());
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, mLinearFiltering ? GL_LINEAR : GL_NEAREST);
	glEnable(GL_CULL_FACE);
	glBegin(GL_TRIANGLES);
	glTexCoord2f(aS0 * aTexture.GetXRatio(), aT0 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV0.X, aV0.Y, aV0.Z);
	glTexCoord2f(aS1 * aTexture.GetXRatio(), aT1 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV1.X, aV1.Y, aV1.Z);
	glTexCoord2f(aS2 * aTexture.GetXRatio(), aT2 * aTexture.GetYRatio());
	glColor3f(1, 1, 1);
	glVertex3f(aV2.X, aV2.Y, aV2.Z);
	glEnd();
	Assert(glGetError() == GL_NO_ERROR)
}



void CApplication::DrawText(CString aText, int aX, int aY)
{
	int X = aX;
	for (int i = 0; aText.HasNext(i); aText.Next(i))
	{
		CTexture& Texture = mFont[CString(aText.GetChar(i), 1)];
		DrawSprite(Texture, X, aY);
		X += Texture.GetXSize();
	}
}



void CApplication::LoadSound(CString aName)
{
	CString FilePath = GetDataFilePath(aName);
	HANDLE File = CreateFile(FilePath.GetData(), GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (File == INVALID_HANDLE_VALUE)
		throw CObeliskException(SL("Sound file not found: ") + aName);
	LARGE_INTEGER Size;
	bool FSResult = GetFileSizeEx(File, &Size);
	Assert(FSResult)
	if (Size.HighPart != 0)
		throw CObeliskException(SL("Sound file too large (> 4GiB): ") + aName);
	TByte *Data = new TByte [Size.LowPart];
	DWORD BytesRead;
	if (!ReadFile(File, Data, Size.LowPart, &BytesRead, NULL))
	{
		Assert(GetLastError() == ERROR_OPERATION_ABORTED)
		delete [] Data;
	}
	else
	{
		Assert(BytesRead == Size.LowPart)
		mSounds.Add(aName, Data);
	}
	bool CHResult = CloseHandle(File);
	Assert(CHResult)
}



void CApplication::Sound(CString aName)
{
	::PlaySound((LPCTSTR)mSounds[aName], NULL, SND_MEMORY | SND_ASYNC | SND_NODEFAULT);
}



IReader* CApplication::OpenDataFile(CString aFileName)
{
	CString FilePath = GetDataFilePath(aFileName);
	HANDLE File = CreateFile(FilePath.GetData(), GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (File == INVALID_HANDLE_VALUE)
		throw CObeliskException(SL("Data file not found: ") + aFileName);
	return new CFileWrapper(File, GetCurrentThreadId() != mMainThreadID);
}



IReader* CApplication::ReadUserFile(CString aFileName)
{
	CString FilePath;
	if (!GetUserFilePath(aFileName, false, FilePath))
		return NULL;
	HANDLE File = CreateFile(FilePath.GetData(), GENERIC_READ, FILE_SHARE_READ, NULL,
			OPEN_EXISTING, FILE_FLAG_SEQUENTIAL_SCAN, NULL);
	if (File == INVALID_HANDLE_VALUE)
		return NULL;
	return new CFileWrapper(File, GetCurrentThreadId() != mMainThreadID);
}



IWriter* CApplication::WriteUserFile(CString aFileName)
{
	CString FilePath;
	if (!GetUserFilePath(aFileName, true, FilePath))
		return NULL;
	HANDLE File = CreateFile(FilePath.GetData(), GENERIC_WRITE, 0, NULL, CREATE_ALWAYS,
			FILE_ATTRIBUTE_NORMAL, NULL);
	if (File == INVALID_HANDLE_VALUE)
		return NULL;
	return new CFileWrapper(File, GetCurrentThreadId() != mMainThreadID);
}