﻿#include <windows.h>
#include <gl\gl.h>

#include "Shared.h"
#include "..\Shared\CObeliskException.h"
#include "CApplication.h"
#include "CThreadWrapper.h"

using namespace Obelisk;



CThreadWrapper::CThreadWrapper(CApplication* aApplication, IThread* aThread)
{
	mApplication = aApplication;
	mThread = aThread;
	mThreadHandle = NULL;
	mStopped = CreateEvent(NULL, true, false, NULL);
	Assert(mStopped != NULL)
	mIOStopped = false;
}



CThreadWrapper::~CThreadWrapper()
{
	#ifndef NO_THREADS
		Assert(mThreadHandle != NULL)
		bool CHResult = CloseHandle(mThreadHandle);
		Assert(CHResult)
	#endif
	CHResult = CloseHandle(mStopped);
	Assert(CHResult)
	delete mThread;
}



void CThreadWrapper::Run()
{
	#ifndef NO_THREADS
		mThreadHandle = CreateThread(NULL, 0, ThreadProc, this, 0, NULL);
		Assert(mThreadHandle != NULL)
	#else
		mThread->Run(this);
		bool PMResult = PostMessage(mApplication->GetWindow(), WM_USER + 1, (WPARAM)mThread, 0);
		Assert(PMResult)
	#endif
}



DWORD WINAPI CThreadWrapper::ThreadProc(LPVOID lpParameter)
{
	((CThreadWrapper*)lpParameter)->ThreadMain();
	return 0;
}



void CThreadWrapper::ThreadMain()
{
	try
	{
		mThread->Run(this);
		if (wglGetCurrentContext() != NULL)
		{
			bool MCResult = wglMakeCurrent(NULL, NULL);
			Assert(MCResult)
		}
		bool PMResult = PostMessage(mApplication->GetWindow(), WM_USER + 1, (WPARAM)mThread, 0);
		Assert(PMResult)
	}
	catch (CObeliskException& e)
	{
		wchar_t *Message = new wchar_t [e.GetMessage().GetSize() + 1];
		wcscpy_s(Message, e.GetMessage().GetSize() + 1, e.GetMessage().GetData());
		PostMessage(mApplication->GetWindow(), WM_USER + 3, (WPARAM)Message, 0);
	}
	catch (CException&)
	{
		PostMessage(mApplication->GetWindow(), WM_USER + 2, 0, 0);
	}
}



void CThreadWrapper::Stop()
{
	bool SEResult = SetEvent(mStopped);
	Assert(SEResult)
	mIOStopped = mApplication->CancelSynchronousIo(mThreadHandle);
	Assert(mIOStopped || GetLastError() == ERROR_NOT_FOUND)
}



bool CThreadWrapper::IsStopped()
{
	DWORD WFSOResult = WaitForSingleObject(mStopped, 0);
	Assert(WFSOResult == WAIT_OBJECT_0 || WFSOResult == WAIT_TIMEOUT)
	if (WFSOResult == WAIT_OBJECT_0)
	{
		mIOStopped = true;
		return true;
	}
	return false;
}



CString CThreadWrapper::IntToString(int aNumber, int aLeading0s)
{
	return mApplication->IntToString(aNumber, aLeading0s);
}



void CThreadWrapper::NotifyWindow(int aMessage, int aData)
{
	mApplication->NotifyWindow(aMessage, aData);
}



double CThreadWrapper::Random()
{
	return mApplication->Random();
}



int CThreadWrapper::Random(int aMin, int aMax)
{
	return mApplication->Random(aMin, aMax);
}



void CThreadWrapper::AttachRenderingContext()
{
	mApplication->AttachRenderingContext();
}



bool CThreadWrapper::LoadTexture(CString aName, bool aTransparent)
{
	if (mIOStopped)
		throw CException("Thread IO not allowed");
	mApplication->LoadTexture(aName, aTransparent);
	return !IsStopped();
}



bool CThreadWrapper::LoadSound(CString aName)
{
	if (mIOStopped)
		throw CException("Thread IO not allowed");
	mApplication->LoadSound(aName);
	return !IsStopped();
}



IReader* CThreadWrapper::OpenDataFile(CString aFileName)
{
	if (mIOStopped)
		throw CException("Thread IO not allowed");
	return IsStopped() ? NULL : mApplication->OpenDataFile(aFileName);
}



IReader* CThreadWrapper::ReadUserFile(CString aFileName)
{
	if (mIOStopped)
		throw CException("Thread IO not allowed");
	return IsStopped() ? NULL : mApplication->ReadUserFile(aFileName);
}



IWriter* CThreadWrapper::WriteUserFile(CString aFileName)
{
	if (mIOStopped)
		throw CException("Thread IO not allowed");
	return IsStopped() ? NULL : mApplication->WriteUserFile(aFileName);
}